/*
 *  TOPPERS Software
 *      Toyohashi Open Platform for Embedded Real-Time Systems
 *
 *  Copyright (C) 2007-2008 by TAKAGI Nobuhisa
 *
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */
#include <cstring>
#include <algorithm>
#include <iostream>
#include <iterator>
#include "toppers/io.hpp"
#include "toppers/cpp.hpp"
#include "toppers/global.hpp"
#include "toppers/diagnostics.hpp"
#include "toppers/oil/preprocess.hpp"
#include <boost/spirit.hpp>

namespace toppers
{
  namespace oil
  {

    /*!
     *  \brief  ÓIAPI `INCLUDE'̓WJ
     *  \param[in]  in      ̓t@C̓e
     *  \param[out] out     o̓t@C̓e
     *  \param[in]  codeset R[h
     *  \param[out] dependencies  ˑ֌W̊i[iNULL̏ꍇ͊i[Ȃj
     *  \param[out] onces   #pragma onceLqĂt@C̊i[iNULL̏ꍇ͊i[Ȃj
     *
     *  `INCLUDE' ̈́AITRON dl̂̂Ƃ͈قȂAdpň͂܂ #include w߂
     *  `pB
     */
    void expand_include( text const& in, text& out, codeset_t codeset, std::set< std::string >* dependencies, std::set< std::string >* onces )
    {
      typedef text::container::const_iterator const_row_iterator;
      typedef std::string::size_type size_type;
      size_type const npos = std::string::npos;
      const_row_iterator first( in.begin().get_row() ), last( in.end().get_row() );

      out.set_line( first->line.file, first->line.line );
      for ( const_row_iterator iter( first ); iter != last; ++iter )
      {
        std::string const& buf = iter->buf;
        char state = 0;

        for ( size_type i = 0, n = buf.size(); i != n; ++i )
        {
          char c = buf[i];
          if ( c == '\'' || c == '\"' )
          {
            if ( state == 0 )
            {
              state = c;    // ijeJn
            }
            else if ( buf[i - 1] != '\\' ) //  \' ܂ \" ł͂Ȃ
            {
              state = 0;    // ijeI
            }
            else if ( ( codeset == shift_jis ) && ( i >= 2 ) && is_lead< shift_jis >( buf[i - 2] ) )
            {
              state = 0;
            }
            out.push_back( c );
          }
          else if ( state != 0 )  // ije
          {
            out.push_back( c );
          }
          else
          {
            using namespace boost::spirit;
            std::string headername;
            if ( iter == last )
            {
              return;
            }
            text::const_iterator iter2( iter, i );
            if ( iter2 == in.end() )
            {
              break;
            }
            parse_info< text::const_iterator > info
              = parse( iter2, in.end(),
				( str_p( "#include" ) >> *space_p
                            >> '\"' >> ( +( anychar_p - '\"' ) )[ assign( headername ) ] >> '\"'
                           ) );  // sԍ̂ŁAXLbvp[T͎gpȂ
            if ( info.hit )
            {
              std::vector< std::string > include_paths = get_global< std::vector< std::string > >( "include-path" );
              std::string hname = search_include_file( include_paths.begin(), include_paths.end(), headername );
              if ( hname.empty() )  // wb_Ȃ
              {
                error( iter->line, _( "cannot open file `%1%\'" ), headername );
              }
              else if ( onces == 0 || onces->find( hname ) == onces->end() )
              {
                if ( dependencies != 0 )
                {
                  dependencies->insert( hname );
                }
                out.push_back( ' ' ); // _~[}ĂȂƍsԍ
                preprocess( hname, out, codeset, dependencies, onces );  // wb_Ŏw肳ꂽt@Cɑ΂đOċAIɍs
                iter = info.stop.get_row();
                i = info.stop.get_col() - 1;
                if ( iter != last )
                {
                  out.set_line( iter->line.file, iter->line.line );
                }
              }
            }
            else
            {
              if ( iter2.get_col() == 0 )
              {
                info = parse( iter2, in.end(),
                              ( *( space_p - eol_p ) >>
                                  ch_p( '#' ) >> *space_p >> "pragma" >> *space_p
                                    >> "once"
                                    >> *( space_p - eol_p ) >> eol_p ) );
                if ( info.hit )
                {
                  if ( onces != 0 )
                  {
                    onces->insert( iter2.line().file );
                  }
                  out.push_back( ' ' );
                  iter = info.stop.get_row();
                  if ( iter != last )
                  {
                    out.set_line( iter->line.file, iter->line.line );
                  }
                  --iter;   // CNĝł߂B
                  break;
                }
              }
              if ( parse( iter2, in.end(),
				  ( str_p( "#include" ) >> *space_p >> '\"' >> +( anychar_p - ')' ) >> '\"') ).hit )
              {
                warning( iter->line, _( "probably, %1% argument of `%2%\' is illegal" ), _( "1st" ), "INCLUDE" );
              }
              out.push_back( c );
            }
          }
        }
      }
    }

    /*!
     *  \brief  RtBM[Vt@C̑O
     *  \param[in]  file    RtBM[Vt@C
     *  \param[out] result  O̓e
     *  \param[in]  codeset R[h
     *  \param[out] dependencies  ˑ֌W̊i[iNULL̏ꍇ͊i[Ȃj
     *  \param[out] onces   #pragma onceLqĂt@C̊i[iNULL̏ꍇ͊i[Ȃj
     */
    void preprocess( std::string const& file, text& result, codeset_t codeset, std::set< std::string >* dependencies, std::set< std::string >* onces )
    {
      std::string buf;
      read( file, buf );
      if ( buf.empty() )
      {
        return;
      }
      text txt;
      txt.set_line( file, 1 );
      remove_comment( buf.begin(), buf.end(), std::back_inserter( txt ), codeset );
      expand_include( txt, result, codeset, dependencies, onces );
    }

  }
}
